import pathlib
import shutil
import subprocess
import unittest.mock

import tooncher


def test_start_engine():
    process = tooncher.start_engine(
        engine_path=pathlib.Path(shutil.which("printenv")),
        gameserver="gameserver",
        playcookie="cookie",
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert isinstance(process, subprocess.Popen)
    stdout, stderr = process.communicate()
    assert not stderr
    env = stdout.strip().split(b"\n")
    assert b"TTR_GAMESERVER=gameserver" in env
    assert b"TTR_PLAYCOOKIE=cookie" in env


@unittest.mock.patch("os.environ", {"SOME": "VAR", "OTHER": "VALUE"})
def test_start_engine_mac():
    app_support_path = "/Users/me/Library/Application Support"
    with unittest.mock.patch("subprocess.Popen") as popen_mock:
        with unittest.mock.patch("sys.platform", "darwin"):
            tooncher.start_engine(
                engine_path=pathlib.PosixPath(
                    app_support_path + "/Toontown Rewritten/Toontown Rewritten"
                ),
                gameserver="gameserver",
                playcookie="cookie",
                check=True,
            )
    popen_mock.assert_called_once_with(
        args=[
            "/Users/me/Library/Application Support/Toontown Rewritten/Toontown Rewritten"
        ],
        check=True,
        cwd=pathlib.PosixPath(
            "/Users/me/Library/Application Support/Toontown Rewritten"
        ),
        env={
            "SOME": "VAR",
            "OTHER": "VALUE",
            "TTR_GAMESERVER": "gameserver",
            "TTR_PLAYCOOKIE": "cookie",
            "DYLD_LIBRARY_PATH": app_support_path
            + "/Toontown Rewritten/Libraries.bundle",
            "DYLD_FRAMEWORK_PATH": app_support_path + "/Toontown Rewritten/Frameworks",
        },
    )


def test_start_engine_xorg():
    with unittest.mock.patch("subprocess.Popen") as popen_mock:
        with unittest.mock.patch("os.environ", {"XAUTHORITY": "/home/me/.Xauthority"}):
            with unittest.mock.patch("sys.platform", "linux"):
                tooncher.start_engine(
                    engine_path=pathlib.PosixPath("/opt/toontown-rewritter/TTREngine"),
                    gameserver="gameserver.tld",
                    playcookie="cookie123",
                    check=False,
                )
    popen_mock.assert_called_once_with(
        args=["/opt/toontown-rewritter/TTREngine"],
        check=False,
        cwd=pathlib.PosixPath("/opt/toontown-rewritter"),
        env={
            "TTR_GAMESERVER": "gameserver.tld",
            "TTR_PLAYCOOKIE": "cookie123",
            "XAUTHORITY": "/home/me/.Xauthority",
        },
    )



import argparse
import os
import pathlib
import sys

import yaml

import tooncher

if sys.platform == "darwin":
    _TOONTOWN_ENGINE_DEFAULT_PATH = os.path.join(
        os.path.expanduser("~"),
        "Library",
        "Application Support",
        "Toontown Rewritten",
        "Toontown Rewritten",
    )
else:
    _TOONTOWN_ENGINE_DEFAULT_PATH = None


def run(
    username,
    config_path,
    engine_path=None,
    validate_ssl_certs=True,
    cpu_limit_percent=None,
):
    if os.path.exists(config_path):
        with open(config_path) as config_file:
            config = yaml.safe_load(config_file.read())
    else:
        config = {}
    if engine_path is None:
        if "engine_path" in config:
            engine_path = config["engine_path"]
        else:
            engine_path = _TOONTOWN_ENGINE_DEFAULT_PATH
    if engine_path is None:
        raise ValueError(
            "missing path to toontown engine\n"
            + "pass --engine-path, set $TOONCHER_ENGINE_PATH, or add to config file"
        )
    accounts = [a for a in config.get("accounts", []) if a["username"] == username]
    if not accounts:
        raise ValueError("username {!r} was not found in config file".format(username))
    if len(accounts) > 1:
        raise ValueError(
            "multiple entries for username {!r} in config file".format(username)
        )
    tooncher.launch(
        engine_path=pathlib.Path(engine_path),
        username=accounts[0]["username"],
        password=accounts[0]["password"],
        validate_ssl_certs=validate_ssl_certs,
        cpu_limit_percent=cpu_limit_percent,
    )


class _EnvDefaultArgparser(argparse.ArgumentParser):
    def add_argument(self, *args, envvar=None, **kwargs):
        if envvar:
            envvar_value = os.environ.get(envvar, None)
            if envvar_value:
                kwargs["required"] = False
                kwargs["default"] = envvar_value
        super().add_argument(*args, **kwargs)


def _init_argparser():
    argparser = _EnvDefaultArgparser(description=None)
    argparser.add_argument("username")
    argparser.add_argument(
        "--config",
        "-c",
        metavar="path",
        dest="config_path",
        help="path to config file (default: %(default)s)",
        default=os.path.join(os.path.expanduser("~"), ".tooncher"),
    )
    argparser.add_argument(
        "--engine-path",
        "-e",
        metavar="path",
        dest="engine_path",
        envvar="TOONCHER_ENGINE_PATH",
        default=None,
        help="path to toontown engine (overrides path in config file, "
        + "may also be set via env var $TOONCHER_ENGINE_PATH)",
    )
    argparser.add_argument(
        "--no-ssl-cert-validation",
        "-k",
        dest="validate_ssl_certs",
        help="do not validate ssl certificates",
        action="store_false",
    )
    argparser.add_argument(
        "--cpu-limit",
        dest="cpu_limit_percent",
        type=int,
        default=None,
        help="maximally allowed cpu usage in percent"
        + " (requires cpulimit command, default: %(default)s)",
    )
    return argparser


def main() -> None:
    argparser = _init_argparser()
    args = argparser.parse_args()
    run(**vars(args))



import pathlib
import subprocess
from unittest.mock import patch

import pytest
import yaml

# pylint: disable=protected-access
import tooncher._cli


def test_cli_help():
    proc_info = subprocess.run(
        ["tooncher", "--help"],
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert b"optional arguments:" in proc_info.stdout
    assert not proc_info.stderr


@patch("tooncher._cli.run")
@patch("os.environ", {})
def test_engine_path_arg(run_mock):
    with patch("sys.argv", ["", "--engine-path", "/opt/ttr/TTREngine", "username"]):
        tooncher._cli.main()
    run_mock.assert_called_once()
    args, kwargs = run_mock.call_args
    assert not args
    assert kwargs["engine_path"] == "/opt/ttr/TTREngine"


@patch("tooncher._cli.run")
@patch("os.environ", {"TOONCHER_ENGINE_PATH": "/opt/ttr/TTREnvine"})
def test_engine_path_env(run_mock):
    with patch("sys.argv", ["", "username"]):
        tooncher._cli.main()
    run_mock.assert_called_once()
    args, kwargs = run_mock.call_args
    assert not args
    assert kwargs["engine_path"] == "/opt/ttr/TTREnvine"


@patch("tooncher._cli.run")
@patch("os.environ", {"TOONCHER_ENGINE_PATH": "/opt/ttr/TTREnvine"})
def test_engine_path_arg_env(run_mock):
    with patch("sys.argv", ["", "--engine-path", "/opt/ttr/TTREngine", "username"]):
        tooncher._cli.main()
    run_mock.assert_called_once()
    args, kwargs = run_mock.call_args
    assert not args
    assert kwargs["engine_path"] == "/opt/ttr/TTREngine"


@patch("tooncher.launch")
@patch("os.environ", {})
def test_engine_path_config(launch_mock, tmpdir):
    config_file = tmpdir.join("config")
    config_file.write(
        yaml.safe_dump(
            {
                "engine_path": "/opt/conf/TTR",
                "accounts": [{"username": "someone", "password": "secret"}],
            }
        )
    )
    with patch("sys.argv", ["", "--config", config_file.strpath, "someone"]):
        tooncher._cli.main()
    launch_mock.assert_called_once()
    args, kwargs = launch_mock.call_args
    assert not args
    assert kwargs["engine_path"] == pathlib.Path("/opt/conf/TTR")


@patch("tooncher.launch")
@patch("os.environ", {"TOONCHER_ENGINE_PATH": "/opt/ttr/TTREnvine"})
def test_engine_path_env_config(launch_mock, tmpdir):
    config_file = tmpdir.join("config")
    config_file.write(
        yaml.safe_dump(
            {
                "engine_path": "/opt/conf/TTR",
                "accounts": [{"username": "someone", "password": "secret"}],
            }
        )
    )
    with patch("sys.argv", ["", "--config", config_file.strpath, "someone"]):
        tooncher._cli.main()
    launch_mock.assert_called_once()
    args, kwargs = launch_mock.call_args
    assert not args
    assert kwargs["engine_path"] == pathlib.Path("/opt/ttr/TTREnvine")


@patch("tooncher.launch")
def test_account(launch_mock, tmpdir):
    config_file = tmpdir.join("config")
    config_file.write(
        yaml.safe_dump(
            {
                "engine_path": "/opt/conf/TTR",
                "accounts": [
                    {"username": "someone", "password": "secret"},
                    {"username": "toon", "password": "town"},
                ],
            }
        )
    )
    with patch("sys.argv", ["", "--config", config_file.strpath, "toon"]):
        tooncher._cli.main()
    launch_mock.assert_called_once_with(
        engine_path=pathlib.Path("/opt/conf/TTR"),
        username="toon",
        password="town",
        validate_ssl_certs=True,
        cpu_limit_percent=None,
    )


def test_account_duplicate_username(tmpdir):
    config_file = tmpdir.join("config")
    config_file.write(
        yaml.safe_dump(
            {
                "engine_path": "/opt/conf/TTR",
                "accounts": [
                    {"username": "someone", "password": "secret"},
                    {"username": "toon", "password": "town"},
                    {"username": "toon", "password": "town2"},
                ],
            }
        )
    )
    with patch("sys.argv", ["", "--config", config_file.strpath, "toon"]):
        with pytest.raises(ValueError, match=r"multiple .* username"):
            tooncher._cli.main()


def test_account_unknown_username(tmpdir):
    config_file = tmpdir.join("config")
    config_file.write(
        yaml.safe_dump(
            {
                "engine_path": "/opt/conf/TTR",
                "accounts": [
                    {"username": "someone", "password": "secret"},
                    {"username": "toon", "password": "town"},
                ],
            }
        )
    )
    with patch("sys.argv", ["", "--config", config_file.strpath, "player"]):
        with pytest.raises(ValueError, match=r"not found"):
            tooncher._cli.main()



import pathlib

import setuptools

setuptools.setup(
    name="tooncher",
    use_scm_version=True,
    packages=setuptools.find_packages(),
    description="automates toontown rewritten's login process",
    long_description=pathlib.Path(__file__).parent.joinpath("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="Fabian Peter Hammerle",
    author_email="fabian.hammerle@gmail.com",
    url="https://github.com/fphammerle/tooncher",
    keywords=["game", "launcher", "toontown rewritten", "ttr"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Topic :: Games/Entertainment",
        "Topic :: Utilities",
    ],
    entry_points={"console_scripts": ["tooncher = tooncher._cli:main"]},
    install_requires=["pyyaml"],
    setup_requires=["setuptools_scm"],
    tests_require=["pytest"],
)



import datetime
import json
import os
import pathlib
import ssl
import subprocess
import sys
import typing
import urllib.parse
import urllib.request

# official api documentation:
# https://github.com/ToontownRewritten/api-doc/blob/master/login.md
# https://github.com/ToontownRewritten/api-doc/blob/master/invasions.md

_LOGIN_API_URL = "https://www.toontownrewritten.com/api/login?format=json"


def start_engine(
    engine_path: pathlib.Path, gameserver: str, playcookie: str, **popen_kwargs
) -> subprocess.Popen:
    # without XAUTHORITY:
    # > :display:x11display(error): Could not open display ":0.0".
    # > :ToonBase: Default graphics pipe is glxGraphicsPipe (OpenGL).
    # > :ToonBase(warning): Unable to open 'onscreen' window.
    # > Traceback (most recent call last):
    # >   File "<compiled '__voltorbmain__'>", line 0, in <module>
    # >   [...]
    # >   File "<compiled 'direct.vlt8f63e471.ShowBase'>", line 0, in vltf05fd21b
    # > Exception: Could not open window.
    # optirun sets plenty of env vars
    env = os.environ.copy()
    env["TTR_GAMESERVER"] = gameserver
    env["TTR_PLAYCOOKIE"] = playcookie
    engine_path = engine_path.resolve()
    if sys.platform == "darwin":
        env["DYLD_LIBRARY_PATH"] = str(engine_path.parent.joinpath("Libraries.bundle"))
        env["DYLD_FRAMEWORK_PATH"] = str(engine_path.parent.joinpath("Frameworks"))
    return subprocess.Popen(
        args=[str(engine_path)], cwd=engine_path.parent, env=env, **popen_kwargs,
    )


def _api_request(
    url: str, params: typing.Optional[dict] = None, validate_ssl_cert: bool = True
):
    resp = urllib.request.urlopen(
        url=url,
        data=urllib.parse.urlencode(params).encode("ascii") if params else None,
        context=None if validate_ssl_cert else ssl._create_unverified_context(),
    )
    return json.loads(resp.read().decode("ascii"))


class _LoginSuccessful:
    def __init__(self, playcookie: str, gameserver: str):
        self.playcookie = playcookie
        self.gameserver = gameserver


class _LoginDelayed:
    def __init__(self, queue_token: str):
        self.queue_token = queue_token


def _login(
    username: typing.Optional[str] = None,
    password: typing.Optional[str] = None,
    queue_token: typing.Optional[str] = None,
    validate_ssl_cert: bool = True,
) -> typing.Union[_LoginSuccessful, _LoginDelayed]:
    if username is not None and queue_token is None:
        assert password is not None
        req_params = {
            "username": username,
            "password": password,
        }
    elif username is None and queue_token is not None:
        req_params = {
            "queueToken": queue_token,
        }
    else:
        raise Exception("either specify username or queue token")
    resp_data = _api_request(
        url=_LOGIN_API_URL, params=req_params, validate_ssl_cert=validate_ssl_cert,
    )
    if resp_data["success"] == "true":
        return _LoginSuccessful(
            playcookie=resp_data["cookie"], gameserver=resp_data["gameserver"],
        )
    if resp_data["success"] == "delayed":
        return _LoginDelayed(queue_token=resp_data["queueToken"],)
    raise Exception(repr(resp_data))


def launch(
    engine_path: pathlib.Path,
    username: str,
    password: str,
    validate_ssl_certs: bool = True,
    cpu_limit_percent: typing.Optional[int] = None,
) -> None:
    result = _login(
        username=username, password=password, validate_ssl_cert=validate_ssl_certs,
    )
    if isinstance(result, _LoginDelayed):
        result = _login(
            queue_token=result.queue_token, validate_ssl_cert=validate_ssl_certs,
        )
    if not isinstance(result, _LoginSuccessful):
        raise Exception("unexpected response: {!r}".format(result))
    process = start_engine(
        engine_path=engine_path,
        gameserver=result.gameserver,
        playcookie=result.playcookie,
    )
    if cpu_limit_percent is not None:
        subprocess.Popen(
            args=[
                "cpulimit",
                "--pid",
                str(process.pid),
                "--limit",
                str(cpu_limit_percent),
                # '--verbose',
            ]
        )
    process.wait()
